# Copyright 2017 DT42
#
# This file is part of BerryNet.
#
# BerryNet is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BerryNet is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with BerryNet.  If not, see <http://www.gnu.org/licenses/>.

"""Movidius inference engine.
"""

from __future__ import print_function

import argparse
import logging

import movidius as mv
from engineservice import EngineService
from engineservice import DLEngine
from dlmodelmgr import DLModelManager


class MovidiusEngine(DLEngine):
    def __init__(self, model, label):
        super(MovidiusEngine, self).__init__()
        self.mvng = mv.MovidiusNeuralGraph(model, label)

    def process_input(self, tensor):
        return mv.process_inceptionv3_input(tensor)

    def inference(self, tensor):
        return self.mvng.inference(tensor)

    def process_output(self, output):
        return mv.process_inceptionv3_output(
                   output,
                   self.mvng.get_labels())

    def save_cache(self):
        with open(self.cache['model_output_filepath'], 'w') as f:
            for i in self.cache['model_output']:
                print("%s (score = %.5f)" % (i[0], i[1]), file=f)


def parse_args():
    ap = argparse.ArgumentParser()
    ap.add_argument('--model',
                    help='Model file path')
    ap.add_argument('--label',
                    help='Label file path')
    ap.add_argument('--model_package',
                    default='',
                    help='Model package "name-version" naming')
    ap.add_argument('--image_dir', required=True,
                    help='Path to image file')
    ap.add_argument('--service_name', required=True,
                    help='Engine service name used as PID filename')
    ap.add_argument('--num_top_predictions', default=5,
                    help='Display this many predictions')
    return vars(ap.parse_args())


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    args = parse_args()
    if args['model_package'] != '':
        dlmm = DLModelManager()
        meta = dlmm.get_model_meta(args['model_package'])
        args['model'] = meta['model']
        args['label'] = meta['label']
    logging.debug('model filepath: ' + args['model'])
    logging.debug('label filepath: ' + args['label'])
    logging.debug('image_dir: ' + args['image_dir'])

    movidius_engine = MovidiusEngine(args['model'], args['label'])
    engine_service = EngineService(args['service_name'], movidius_engine)
    engine_service.run(args)



import logging
import time

from argparse import ArgumentParser
from os import path

import cv2
import numpy as np
import tensorflow as tf

from berrynet.engine import DLEngine
from berrynet import logger


class TFLiteYoloV4DetectorEngine(DLEngine):
    def __init__(self, model, labels, threshold=0.1, num_threads=1):
        """
        Builds Tensorflow graph, load model and labels
        """
        # Load labels
        self.labels = self._load_label(labels)
        self.classes = len(self.labels)

        # Define lite graph and Load Tensorflow Lite model into memory
        self.interpreter = tf.lite.Interpreter(
            model_path=model)
        self.interpreter.allocate_tensors()
        self.input_details = self.interpreter.get_input_details()
        self.output_details = self.interpreter.get_output_details()
        self.input_dtype = self.input_details[0]['dtype']
        self.num_threads = num_threads
        self.threshold = threshold

    def __delete__(self, instance):
        #tf.reset_default_graph()
        #self.sess = tf.InteractiveSession()
        del self.interpreter

    def process_input(self, tensor):
        """Resize and normalize image for network input"""

        self.img_w = tensor.shape[1]
        self.img_h = tensor.shape[0]

        frame = cv2.cvtColor(tensor, cv2.COLOR_BGR2RGB)
        frame = cv2.resize(frame, (416, 416))
        frame = np.expand_dims(frame, axis=0)
        if self.input_dtype == np.float32:
            frame = (1.0 / 255.0) * frame
            frame = frame.astype('float32')
        else:
            # default data type returned by cv2.imread is np.unit8
            pass
        return frame

    def filter_boxes(self, box_xywh, scores, score_threshold=0.4, input_shape = tf.constant([416,416])):
        scores_max = tf.math.reduce_max(scores, axis=-1)

        mask = scores_max >= score_threshold
        class_boxes = tf.boolean_mask(box_xywh, mask)
        pred_conf = tf.boolean_mask(scores, mask)
        class_boxes = tf.reshape(class_boxes, [tf.shape(scores)[0], -1, tf.shape(class_boxes)[-1]])
        pred_conf = tf.reshape(pred_conf, [tf.shape(scores)[0], -1, tf.shape(pred_conf)[-1]])

        box_xy, box_wh = tf.split(class_boxes, (2, 2), axis=-1)

        input_shape = tf.cast(input_shape, dtype=tf.float32)

        box_yx = box_xy[..., ::-1]
        box_hw = box_wh[..., ::-1]

        box_mins = (box_yx - (box_hw / 2.)) / input_shape
        box_maxes = (box_yx + (box_hw / 2.)) / input_shape
        boxes = tf.concat([
            box_mins[..., 0:1],  # y_min
            box_mins[..., 1:2],  # x_min
            box_maxes[..., 0:1],  # y_max
            box_maxes[..., 1:2]  # x_max
        ], axis=-1)
        # return tf.concat([boxes, pred_conf], axis=-1)
        return (boxes, pred_conf)
    
    def inference(self, tensor):
        #self.interpreter.set_num_threads(int(self.num_threads));
        self.interpreter.set_tensor(self.input_details[0]['index'], tensor)
        self.interpreter.invoke()

        # get results
        pred = [self.interpreter.get_tensor(self.output_details[i]['index']) for i in range(len(self.output_details))]

        boxes, pred_conf = self.filter_boxes(pred[0], pred[1], score_threshold=0.25, input_shape=tf.constant([416, 416]))
        
        boxes, scores, classes, valid_detections = tf.image.combined_non_max_suppression(
            boxes=tf.reshape(boxes, (tf.shape(boxes)[0], -1, 1, 4)),
            scores=tf.reshape(
                pred_conf, (tf.shape(pred_conf)[0], -1, tf.shape(pred_conf)[-1])),
            max_output_size_per_class=50,
            max_total_size=50,
            iou_threshold=0.45,
            score_threshold=0.25
        )


        return {
            'boxes': boxes,
            'classes': classes,
            'scores': scores,
            'num': valid_detections
        }

    def process_output(self, output):
        # get results
        boxes = np.squeeze(output['boxes'][0])
        classes = np.squeeze(output['classes'][0]).astype(np.int32)
        scores = np.squeeze(output['scores'][0])
        num = output['num'][0]

        annotations = []
        number_boxes = boxes.shape[0]
        for i in range(number_boxes):
            box = tuple(boxes[i].tolist())
            ymin, xmin, ymax, xmax = box

            if scores[i] < self.threshold:
                continue
            annotations.append({
                'label': self.labels[classes[i]],
                'confidence': float(scores[i]),
                'left': int(xmin * self.img_w),
                'top': int(ymin * self.img_h),
                'right': int(xmax * self.img_w),
                'bottom': int(ymax * self.img_h)
            })
        return {'annotations': annotations}

    def _load_label(self, path):
        with open(path, 'r') as f:
            labels = list(map(str.strip, f.readlines()))
        return labels

def parse_argsr():
    parser = ArgumentParser()
    parser.add_argument(
        "-e", "--engine",
        help=("Classifier or Detector engine. "
              "classifier, or detector is acceptable. "
              "(detector by default)"),
        default="detector",
        type=str)
    parser.add_argument(
        "-m", "--model",
        help="Path to an .xml file with a trained model.",
        required=True,
        type=str)
    parser.add_argument(
        "-l", "--labels",
        help="Labels mapping file",
        default=None,
        type=str)
    parser.add_argument(
        "--top_k",
        help="Number of top results",
        default=3,
        type=int)
    parser.add_argument(
        "--num_threads",
        help="Number of threads",
        default=1,
        type=int)
    parser.add_argument(
        "-i", "--input",
        help="Path to a folder with images or path to an image files",
        required=True,
        type=str)
    parser.add_argument(
        "--debug",
        help="Debug mode toggle",
        default=False,
        action="store_true")

    return parser.parse_args()

def main():
    # Example command
    #     $ python3 tfliteyolov4_engine.py -m yolov4-tiny-*.tflite \
    #           --labels coco.names -i dog.jpg --debug
    args = parse_argsr()

    if args.debug:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)

    if args.engine == 'detector':
        engine = TFLiteYoloV4DetectorEngine(
                     model = args.model,
                     labels = args.labels,
                     num_threads = args.num_threads)
    else:
        raise Exception('Illegal engine {}, it should be '
                        'classifier or detector'.format(args.engine))

    for i in range(5):
        bgr_array = cv2.imread(args.input)
        t = time.time()
        image_data = engine.process_input(bgr_array)
        output = engine.inference(image_data)
        model_outputs = engine.process_output(output)
        # Reference result
        #     input:
        #         darknet/data/dog.jpg
        #     output:
        #         Inference takes 0.7533011436462402 s
        #         Inference takes 0.5741658210754395 s
        #         Inference takes 0.6120760440826416 s
        #         Inference takes 0.6191139221191406 s
        #         Inference takes 0.5809791088104248 s
        #         label: bicycle  conf: 0.9563907980918884  (139, 116) (567, 429)
        #         label: car  conf: 0.8757821917533875  (459, 80) (690, 172)
        #         label: dog  conf: 0.869189441204071  (131, 218) (314, 539)
        #         label: car  conf: 0.40003547072410583  (698, 122) (724, 152)
        logger.debug('Inference takes {} s'.format(time.time() - t))

    if args.engine == 'classifier':
        for r in model_outputs['annotations']:
            logger.debug('label: {0}  conf: {1}'.format(
                r['label'],
                r['confidence']
            ))
    elif args.engine == 'detector':
        for r in model_outputs['annotations']:
            logger.debug('label: {0}  conf: {1}  ({2}, {3}) ({4}, {5})'.format(
                r['label'],
                r['confidence'],
                r['left'],
                r['top'],
                r['right'],
                r['bottom']
            ))
    else:
        raise Exception('Can not get result '
                        'from illegal engine {}'.format(args.engine))


if __name__ == '__main__':
    main()



# Copyright 2017 DT42
#
# This file is part of BerryNet.
#
# BerryNet is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BerryNet is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with BerryNet.  If not, see <http://www.gnu.org/licenses/>.

"""Engine service is a bridge between incoming data and inference engine.
"""


from __future__ import print_function

import argparse
import logging
import multiprocessing
import os
import signal
import sys
import threading
import time

import cv2
import numpy as np
import Queue

from dlmodelmgr import DLModelManager
from watchdog.observers import Observer
from watchdog.events import PatternMatchingEventHandler


class DLEngine(object):
    def __init__(self):
        self.model_input_cache = []
        self.model_output_cache = []
        self.cache = {
            'model_input': [],
            'model_output': '',
            'model_output_filepath': ''
        }

    def create(self):
        # Workaround to posepone TensorFlow initialization.
        # If TF is initialized in __init__, and pass an engine instance
        # to engine service, TF session will stuck in run().
        pass

    def process_input(self, tensor):
        return tensor

    def inference(self, tensor):
        output = None
        return output

    def process_output(self, output):
        return output

    def cache_data(self, key, value):
        self.cache[key] = value

    def save_cache(self):
        with open(self.cache['model_output_filepath'], 'w') as f:
            f.write(str(self.cache['model_output']))


class EventHandler(PatternMatchingEventHandler):
    def process(self, event):
        """
        event.event_type
            'modified' | 'created' | 'moved' | 'deleted'
        event.is_directory
            True | False
        event.src_path
            path/to/observed/file
        """
        # the file will be processed there
        _msg = event.src_path
        self.image_queue.put(_msg.rstrip('.done'))
        os.remove(_msg)
        logging.debug(_msg + ' ' + event.event_type)

    # ignore all other types of events except 'modified'
    def on_created(self, event):
        self.process(event)


class EngineService(object):
    def __init__(self, service_name, engine):
        self.service_name = service_name
        self.engine = engine
        self.image_queue = Queue.Queue()
        self.event_handler = EventHandler(['*.jpg.done'])
        # NOTE: Increase object reference count (share memory)
        #       instead of object creation.
        self.event_handler.image_queue = self.image_queue

    def touch(self, fname, times=None):
        with open(fname, 'a'):
            os.utime(fname, times)

    def server(self):
        """Infinite loop serving inference requests"""

        logging.info(threading.current_thread().getName() + " is running")

        self.engine.create()
        while True:
            input_name = self.image_queue.get()
            image_data = cv2.imread(input_name).astype(np.float32)
            self.engine.cache_data('model_input', image_data)
            image_data = self.engine.process_input(image_data)

            output = self.engine.inference(image_data)
            model_outputs = self.engine.process_output(output)
            self.engine.cache_data('model_output', model_outputs)

            output_name = input_name + '.txt'
            output_done_name = output_name + '.done'
            self.engine.cache_data('model_output_filepath', output_name)
            self.engine.save_cache()
            self.touch(output_done_name)
            logging.debug(input_name + " classified!")

    def run(self, args):
        self.record_pid()

        # workaround the issue that SIGINT cannot be received (fork a child to
        # avoid blocking the main process in Thread.join()
        child_pid = os.fork()
        if child_pid == 0:  # child
            # observer handles event in a different thread
            observer = Observer()
            observer.schedule(self.event_handler, path=args['image_dir'])
            observer.start()
            self.server()
        else:  # parent
            try:
                os.wait()
            except KeyboardInterrupt:
                os.kill(child_pid, signal.SIGKILL)
                self.erase_pid()

    def record_pid(self):
        """Write a PID pidfile /tmp/<service_name>.pid.
        """
        pid = str(os.getpid())
        pidfile = '/tmp/{}.pid'.format(self.service_name)
        if os.path.isfile(pidfile):
            logging.critical("%s already exists, exiting" % pidfile)
            sys.exit(1)
        with open(pidfile, 'w') as f:
            f.write(pid)

    def erase_pid(self):
        pidfile = '/tmp/{}.pid'.format(self.service_name)
        os.unlink(pidfile)


def parse_args():
    ap = argparse.ArgumentParser()
    ap.add_argument('--model',
                    help='Model file path')
    ap.add_argument('--label',
                    help='Label file path')
    ap.add_argument('--model_package',
                    default='',
                    help='Model package name')
    ap.add_argument('--image_dir', required=True,
                    help='Path to image file')
    ap.add_argument('--service_name', required=True,
                    help='Engine service name used as PID filename')
    ap.add_argument('--num_top_predictions', default=5,
                    help='Display this many predictions')
    return vars(ap.parse_args())


if __name__ == '__main__':
    import movidius as mv

    logging.basicConfig(level=logging.DEBUG)
    args = parse_args()
    if args['model_package'] != '':
        dlmm = DLModelManager()
        meta = dlmm.get_model_meta(args['model_package'])
        args['model'] = meta['model']
        args['label'] = meta['label']
    logging.debug('model filepath: ' + args['model'])
    logging.debug('label filepath: ' + args['label'])
    logging.debug('image_dir: ' + args['image_dir'])

    mvng = mv.MovidiusNeuralGraph(args['model'], args['label'])
    engine_service = EngineService(args['service_name'], mvng)
    engine_service.run(args)



# Copyright 2018 DT42
#
# This file is part of BerryNet.
#
# BerryNet is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BerryNet is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with BerryNet.  If not, see <http://www.gnu.org/licenses/>.

"""Framebuffer dashboard.
"""

import argparse
import json
import logging
import os
import random
import sys
import time

from datetime import datetime
from os.path import join as pjoin

import cv2

from berrynet import logger
from berrynet.comm import Communicator
from berrynet.comm import payload
from OpenGL.GL import *
from OpenGL.GLU import *
from OpenGL.GLUT import *


class FBDashboardService(object):
    def __init__(self, comm_config, data_dirpath=None, no_decoration=False,
                 debug=False, save_frame=False):
        self.comm_config = comm_config
        for topic, functor in self.comm_config['subscribe'].items():
            self.comm_config['subscribe'][topic] = eval(functor)
        self.comm = Communicator(self.comm_config, debug=True)
        self.data_dirpath = data_dirpath
        self.no_decoration = no_decoration
        self.frame = None
        self.debug = debug
        self.save_frame = save_frame

    def update(self, pl):
        payload_json = payload.deserialize_payload(pl.decode('utf-8'))
        if 'bytes' in payload_json.keys():
            img_k = 'bytes'
        elif 'image_blob' in payload_json.keys():
            img_k = 'image_blob'
        else:
            raise Exception('No image data in MQTT payload')
        jpg_bytes = payload.destringify_jpg(payload_json[img_k])
        payload_json.pop(img_k)
        logger.debug('inference text result: {}'.format(payload_json))

        img = payload.jpg2rgb(jpg_bytes)

        if self.no_decoration:
            self.frame = img
        else:
            try:
                res = payload_json['annotations']
            except KeyError:
                res = [
                    {
                        'label': 'hello',
                        'confidence': 0.42,
                        'left': random.randint(50, 60),
                        'top': random.randint(50, 60),
                        'right': random.randint(300, 400),
                        'bottom': random.randint(300, 400)
                    }
                ]
            self.frame = overlay_on_image(img, res)

        # Save frames for analysis or debugging
        if self.debug and self.save_frame:
            if not os.path.exists(self.data_dirpath):
                try:
                    os.mkdir(self.data_dirpath)
                except Exception as e:
                    logger.warn('Failed to create {}'.format(self.data_dirpath))
                    raise(e)

            timestamp = datetime.now().isoformat()
            with open(pjoin(self.data_dirpath, timestamp + '.jpg'), 'wb') as f:
                f.write(jpg_bytes)
            with open(pjoin(self.data_dirpath, timestamp + '.json'), 'w') as f:
                f.write(json.dumps(payload_json, indent=4))

    def update_fb(self):
        if self.frame is not None:
            gl_draw_fbimage(self.frame)

    def run(self, args):
        """Infinite loop serving inference requests"""
        self.comm.start_nb()


def gl_draw_fbimage(rgbimg):
    h, w = rgbimg.shape[:2]

    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, w, h, 0, GL_RGB, GL_UNSIGNED_BYTE, rgbimg)
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
    glColor3f(1.0, 1.0, 1.0)
    glEnable(GL_TEXTURE_2D)
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR)
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR)
    glBegin(GL_QUADS)
    glTexCoord2d(0.0, 1.0)
    glVertex3d(-1.0, -1.0,  0.0)
    glTexCoord2d(1.0, 1.0)
    glVertex3d( 1.0, -1.0,  0.0)
    glTexCoord2d(1.0, 0.0)
    glVertex3d( 1.0,  1.0,  0.0)
    glTexCoord2d(0.0, 0.0)
    glVertex3d(-1.0,  1.0,  0.0)
    glEnd()
    glFlush()
    glutSwapBuffers()


def init():
    glClearColor(0.7, 0.7, 0.7, 0.7)


def idle():
    glutPostRedisplay()


def keyboard(key, x, y):
    key = key.decode('utf-8')
    if key == 'q':
        print("\n\nFinished\n\n")
        sys.exit()


def opencv_frame(src, w=None, h=None, fps=30):
    vidcap = cv2.VideoCapture(src)
    if not vidcap.isOpened():
        print('opened failed')
        sys.exit(errno.ENOENT)

    # set frame w/h if indicated
    if w and h:
        vidcap.set(cv2.CAP_PROP_FRAME_WIDTH, w)
        vidcap.set(cv2.CAP_PROP_FRAME_HEIGHT, h)

    # set FPS
    rate = int(vidcap.get(cv2.CAP_PROP_FPS))
    if rate > fps or rate < 1:
        print('Illegal data rate {} (1-30)'.format(rate))
        rate = fps
    print('fps: {}'.format(rate))

    # frame generator
    while True:
        success, image = vidcap.read()
        if not success:
            print('Failed to read frame')
            break
        yield image


#Vcap = opencv_frame(0, w=320, h=240)
Vcap = opencv_frame(0)


def draw_box(image, annotations):
    """Draw information of annotations onto image.

    Args:
        image: Image nparray.
        annotations: List of detected object information.

    Returns: Image nparray containing object information on it.
    """
    print('draw_box, annotations: {}'.format(annotations))
    img = image.copy()

    for anno in annotations:
        # draw bounding box
        box_color = (0, 0, 255)
        box_thickness = 1
        cv2.rectangle(img,
                      (int(anno['left']), int(anno['top'])),
                      (int(anno['right']), int(anno['bottom'])),
                      box_color,
                      box_thickness)

        # draw label
        label_background_color = box_color
        label_text_color = (255, 255, 255)
        if 'track_id' in anno.keys():
            label = 'ID:{} {}'.format(anno['track_id'], anno['label'])
        else:
            label = anno['label']
        label_text = '{} ({} %)'.format(label,
                                        int(anno['confidence'] * 100))
        label_size = cv2.getTextSize(label_text,
                                     cv2.FONT_HERSHEY_SIMPLEX,
                                     0.5,
                                     1)[0]
        label_left = anno['left']
        label_top = anno['top'] - label_size[1]
        if (label_top < 1):
            label_top = 1
        label_right = label_left + label_size[0]
        label_bottom = label_top + label_size[1]
        cv2.rectangle(img,
                      (int(label_left - 1), int(label_top - 1)),
                      (int(label_right + 1), int(label_bottom + 1)),
                      label_background_color,
                      -1)
        cv2.putText(img,
                    label_text,
                    (int(label_left), int(label_bottom)),
                    cv2.FONT_HERSHEY_SIMPLEX,
                    0.5,
                    label_text_color,
                    1)
    return img


def overlay_on_image(display_image, object_info):
    """Modulized version of overlay_on_image function
    """
    if isinstance(object_info, type(None)):
        print('WARNING: object info is None')
        return display_image

    return draw_box(display_image, object_info)


def parse_args():
    ap = argparse.ArgumentParser()
    ap.add_argument(
        '--data-dirpath',
        default='/tmp/berrynet-data',
        help='Dirpath where to store collected data.'
    )
    ap.add_argument(
        '--broker-ip',
        default='localhost',
        help='MQTT broker IP.'
    )
    ap.add_argument(
        '--broker-port',
        default=1883,
        type=int,
        help='MQTT broker port.'
    )
    ap.add_argument(
        '--topic',
        nargs='*',
        default=['berrynet/engine/tflitedetector/result'],
        help='The topic to listen, and can be indicated multiple times.'
    )
    ap.add_argument(
        '--topic-action',
        default='self.update',
        help='The action for the indicated topics.'
    )
    ap.add_argument(
        '--topic-config',
        default=None,
        help='Path of the MQTT topic subscription JSON.'
    )
    ap.add_argument(
        '--no-decoration',
        action='store_true',
        help='Display image in payload without applying result information.'
    )
    ap.add_argument(
        '--no-full-screen',
        action='store_true',
        help='Display fbdashboard in a window.'
    )
    ap.add_argument('--debug',
        action='store_true',
        help='Debug mode toggle'
    )
    ap.add_argument('--debug-save-frame',
        action='store_true',
        help='Save frames for debugging. --debug also needs to be set.'
    )
    return vars(ap.parse_args())


def main():
    args = parse_args()
    if args['debug']:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)

    # Topics and actions can come from two sources: CLI and config file.
    # Setup topic_config by parsing values from the two sources.
    if args['topic_config']:
        with open(args['topic_config']) as f:
            topic_config = json.load(f)
    else:
        topic_config = {}
    topic_config.update({t:args['topic_action'] for t in args['topic']})

    comm_config = {
        'subscribe': topic_config,
        'broker': {
            'address': args['broker_ip'],
            'port': args['broker_port']
        }
    }
    fbd_service = FBDashboardService(comm_config,
                                     args['data_dirpath'],
                                     args['no_decoration'],
                                     args['debug'],
                                     args['debug_save_frame'])
    fbd_service.run(args)

    glutInitWindowPosition(0, 0)
    glutInit(sys.argv)
    glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE)
    glutCreateWindow("BerryNet Result Dashboard, q to quit")
    glutDisplayFunc(fbd_service.update_fb)
    glutKeyboardFunc(keyboard)
    init()
    glutIdleFunc(idle)
    if args['no_full_screen']:
        pass
    else:
        glutFullScreen()
    glutMainLoop()


if __name__ == '__main__':
    main()



# Copyright 2018 DT42
#
# This file is part of BerryNet.
#
# BerryNet is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BerryNet is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with BerryNet.  If not, see <http://www.gnu.org/licenses/>.

"""Mockup service with relay engine (default engine).
"""

import argparse
import json
import logging
import os

from berrynet import logger
from berrynet.comm import payload
from berrynet.engine import DLEngine
from berrynet.service import EngineService


class MockupEngine(DLEngine):
    def inference(self, tensor):
        return {
            'annotations': {
                'label': 'dt42',
                'confidence': 0.99
            }
        }


class MockupService(EngineService):
    def __init__(self, service_name, engine, comm_config):
        super().__init__(service_name,
                         engine,
                         comm_config)
        if not os.path.exists('/tmp/mockup'):
            os.mkdir('/tmp/mockup')

    #def generalize_result(self, eng_input, eng_output):
    #    logger.debug()
    #    eng_input.update(eng_output)
    #    return eng_input

    def result_hook(self, generalized_result):
        gr = generalized_result
        jpg_bytes = payload.destringify_jpg(gr.pop('bytes'))
        logger.debug('generalized result (readable only): {}'.format(gr))
        with open('/tmp/mockup/{}.jpg'.format(gr['timestamp']), 'wb') as f:
            f.write(jpg_bytes)
        with open('/tmp/mockup/{}.json'.format(gr['timestamp']), 'w') as f:
            f.write(json.dumps(gr, indent=4))


def parse_args():
    ap = argparse.ArgumentParser()
    ap.add_argument('--debug',
                    action='store_true',
                    help='Debug mode toggle')
    return vars(ap.parse_args())


def main():
    args = parse_args()
    if args['debug']:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)

    eng = MockupEngine()
    comm_config = {
        'subscribe': {},
        'broker': {
            'address': 'localhost',
            'port': 1883
        }
    }
    engine_service = MockupService('mockup service', eng, comm_config)
    engine_service.run(args)


if __name__ == '__main__':
    main()



# Copyright 2017 DT42
#
# This file is part of BerryNet.
#
# BerryNet is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BerryNet is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with BerryNet.  If not, see <http://www.gnu.org/licenses/>.

"""Engine service is a bridge between incoming data and inference engine.
"""

import argparse
import logging

from berrynet import logger
from berrynet.comm import payload
from berrynet.dlmodelmgr import DLModelManager
from berrynet.engine.tensorflow_engine import TensorFlowEngine
from berrynet.service import EngineService


class TensorFlowService(EngineService):
    def __init__(self, service_name, engine, comm_config):
        super(TensorFlowService, self).__init__(service_name,
                                                engine,
                                                comm_config)

    def result_hook(self, generalized_result):
        logger.debug('result_hook, annotations: {}'.format(generalized_result['annotations']))
        self.comm.send('berrynet/engine/tensorflow/result',
                       payload.serialize_payload(generalized_result))


def parse_args():
    ap = argparse.ArgumentParser()
    ap.add_argument('--model',
                    help='Model file path')
    ap.add_argument('--label',
                    help='Label file path')
    ap.add_argument('--model_package',
                    default='',
                    help='Model package name. Find model and label file paths automatically.')
    ap.add_argument('--service_name',
                    default='tensorflow',
                    help='Human-readable service name for service management.')
    ap.add_argument('--num_top_predictions',
                    help='Display this many predictions',
                    default=3,
                    type=int)
    ap.add_argument('--debug',
                    action='store_true',
                    help='Debug mode toggle')
    return vars(ap.parse_args())


def main():
    # Test TensorFlow engine
    args = parse_args()
    if args['debug']:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)
    logger.debug('model filepath: ' + args['model'])
    logger.debug('label filepath: ' + args['label'])

    model = 'berrynet/engine/inception_v3_2016_08_28_frozen.pb'
    label = 'berrynet/engine/imagenet_slim_labels.txt'
    jpg_filepath = 'berrynet/engine/grace_hopper.jpg'
    input_layer = 'input:0'
    output_layer = 'InceptionV3/Predictions/Reshape_1:0'

    tfe = TensorFlowEngine(model, label, input_layer, output_layer)
    comm_config = {
        'subscribe': {},
        'broker': {
            'address': 'localhost',
            'port': 1883
        }
    }
    engine_service = TensorFlowService(args['service_name'],
                                       tfe,
                                       comm_config)
    engine_service.run(args)


if __name__ == '__main__':
    # Test Movidius engine
    #import movidius as mv

    #logging.basicConfig(level=logging.DEBUG)
    #args = parse_args()
    #if args['model_package'] != '':
    #    dlmm = DLModelManager()
    #    meta = dlmm.get_model_meta(args['model_package'])
    #    args['model'] = meta['model']
    #    args['label'] = meta['label']
    #logging.debug('model filepath: ' + args['model'])
    #logging.debug('label filepath: ' + args['label'])
    #logging.debug('image_dir: ' + args['image_dir'])

    #mvng = mv.MovidiusNeuralGraph(args['model'], args['label'])
    #engine_service = EngineService(args['service_name'], mvng)
    #engine_service.run(args)

    main()



# Copyright 2018 DT42
#
# This file is part of BerryNet.
#
# BerryNet is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BerryNet is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with BerryNet.  If not, see <http://www.gnu.org/licenses/>.

"""Utility Functions.
"""

import math

import cv2

from berrynet.comm import payload


def generate_class_color(class_num=20):
    """Generate a RGB color set based on given class number.

    Args:
        class_num: Default is VOC dataset class number.

    Returns:
        A tuple containing RGB colors.
    """
    colors = [(1, 0, 1), (0, 0, 1), (0, 1, 1),
              (0, 1, 0), (1, 1, 0), (1, 0, 0)]
    const = 1234567  # only for offset calculation

    colorset = []
    for cls_i in range(class_num):
        offset = cls_i * const % class_num

        ratio = (float(offset) / class_num) * (len(colors) - 1)
        i = math.floor(ratio)
        j = math.ceil(ratio)
        ratio -= i

        rgb = []
        for ch_i in range(3):
            r = (1 - ratio) * colors[i][ch_i] + ratio * colors[j][ch_i]
            rgb.append(math.ceil(r * 255))
        colorset.append(tuple(rgb[::-1]))
    return tuple(colorset)


def draw_bb(bgr_nparr, infres, class_colors, labels):
    """Draw bounding boxes on an image.

    Args:
        bgr_nparr: image data in numpy array format
        infres: Darkflow inference results
        class_colors: Bounding box color candidates, list of RGB tuples.

    Returens:
        Generalized result whose image data is drew w/ bounding boxes.
    """
    for res in infres['annotations']:
        left = int(res['left'])
        top = int(res['top'])
        right = int(res['right'])
        bottom = int(res['bottom'])
        label = res['label']
        color = class_colors[labels.index(label)]
        confidence = res['confidence']
        imgHeight, imgWidth, _ = bgr_nparr.shape
        thick = int((imgHeight + imgWidth) // 300)

        cv2.rectangle(bgr_nparr,(left, top), (right, bottom), color, thick)
        cv2.putText(bgr_nparr, label, (left, top - 12), 0, 1e-3 * imgHeight,
            color, thick//3)
    #cv2.imwrite('prediction.jpg', bgr_nparr)
    infres['bytes'] = payload.stringify_jpg(
                                    cv2.imencode('.jpg', bgr_nparr)[1])
    return infres


def draw_box(image, annotations):
    """Draw information of annotations onto image.

    Args:
        image: Image nparray.
        annotations: List of detected object information.

    Returns: Image nparray containing object information on it.
    """
    print('draw_box, annotations: {}'.format(annotations))
    img = image.copy()

    for anno in annotations:
        # draw bounding box
        box_color = (0, 0, 255)
        box_thickness = 1
        cv2.rectangle(img,
                      (anno['left'], anno['top']),
                      (anno['right'], anno['bottom']),
                      box_color,
                      box_thickness)

        # draw label
        label_background_color = box_color
        label_text_color = (255, 255, 255)
        if 'track_id' in anno.keys():
            label = 'ID:{} {}'.format(anno['track_id'], anno['label'])
        else:
            label = anno['label']
        label_text = '{} ({} %)'.format(label,
                                        int(anno['confidence'] * 100))
        label_size = cv2.getTextSize(label_text,
                                     cv2.FONT_HERSHEY_SIMPLEX,
                                     0.5,
                                     1)[0]
        label_left = anno['left']
        label_top = anno['top'] - label_size[1]
        if (label_top < 1):
            label_top = 1
        label_right = label_left + label_size[0]
        label_bottom = label_top + label_size[1]
        cv2.rectangle(img,
                      (label_left - 1, label_top - 1),
                      (label_right + 1, label_bottom + 1),
                      label_background_color,
                      -1)
        cv2.putText(img,
                    label_text,
                    (label_left, label_bottom),
                    cv2.FONT_HERSHEY_SIMPLEX,
                    0.5,
                    label_text_color,
                    1)
    return img


def overlay_on_image(display_image, object_info):
    """Modulized version of overlay_on_image function
    """
    if isinstance(object_info, type(None)):
        print('WARNING: object info is None')
        return display_image

    return draw_box(display_image, object_info)


def draw_label(bgr_nparr, infres, class_color, save_image_path=None):
    """Draw bounding boxes on an image.

    Args:
        bgr_nparr: image data in numpy array format
        infres: Inference results followed generic format specification.
        class_color: Label color, a RGB tuple.

    Returens:
        Generalized result whose image data is drew w/ labels.
    """
    left = 0
    top = 0
    for res in infres['annotations']:
        imgHeight, imgWidth, _ = bgr_nparr.shape
        thick = int((imgHeight + imgWidth) // 300)

        # putText can not handle newline char yet,
        # so we have to put multiple texts manually.
        cv2.putText(bgr_nparr,
                    '{0}: {1}'.format(res['label'], res['confidence']),
                    (left + 10, top + 20),  # bottom-left corner of text
                    0,                      # fontFace
                    1e-3 * imgHeight,       # fontScale
                    class_color,
                    thick // 3)
        top += 20
    infres['bytes'] = payload.stringify_jpg(
                                    cv2.imencode('.jpg', bgr_nparr)[1])

    if save_image_path:
        cv2.imwrite(save_image_path, bgr_nparr)

    return infres



#!/usr/bin/env python3
#
# Copyright 2018 DT42
#
# This file is part of BerryNet.
#
# BerryNet is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BerryNet is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with BerryNet.  If not, see <http://www.gnu.org/licenses/>.

"""Simple image classification server with Inception.

The server monitors image_dir and run inferences on new images added to the
directory. Every image file should come with another empty file with '.done'
suffix to signal readiness. Inference result of a image can be read from the
'.txt' file of that image after '.txt.done' is spotted.

This is an example the server expects clients to do. Note the order.

# cp cat.jpg /run/image_dir
# touch /run/image_dir/cat.jpg.done

Clients should wait for appearance of 'cat.jpg.txt.done' before getting
result from 'cat.jpg.txt'.
"""


from __future__ import print_function
import os
import sys
import time
import numpy as np
import threading
import multiprocessing
import queue
import signal
from watchdog.observers import Observer
from watchdog.events import PatternMatchingEventHandler
import caffe
import hashlib
import urllib.request
import tempfile
import shutil

image_queue = queue.Queue()
sess = None
threads = []
image_dir = '/run/image_dir'
caffe_classifier = None
caffe_labels = []
model_meta_file = '/usr/share/doc/caffe-doc/models/bvlc_reference_caffenet/readme.md'
label_file = '/tmp/synset_words.txt'
pretrained_model = None

def logging(*args):
    print("[%08.3f]" % time.time(), ' '.join(args))

def touch(fname, times=None):
    with open(fname, 'a'):
        os.utime(fname, times)

def load_labels(filename):
    """Read in labels, one label per line."""
    return [line.rstrip() for line in open(filename)]

def read_model_meta_file(meta_file):
    """Read model meta file. The meta file is inside caffe-doc package"""
    # We believe we shouldn't read this file for downloading and checking
    # model. Instead we should package some model if there is free one.
    url = None
    sha1sum = None
    filename = None
    for line in open(meta_file):
        l1 = line.rstrip()
        if (l1.startswith("sha1:")):
            sha1sum = l1[len("sha1:"):].strip()
        if (l1.startswith("caffemodel_url:")):
            url = l1[len("caffemodel_url:"):].strip()
        if (l1.startswith("caffemodel:")):
            filename = l1[len("caffemodel:"):].strip()
        if ((sha1sum != None) and (url != None) and (filename != None)):
            break
    if ((url != None) and (sha1sum != None) and (filename != None)):
        return {'url': url, 'sha1sum': sha1sum, 'filename': filename}
    return None

def sha1sum(filename):
    """calculate sha1sum"""
    BUF_SIZE=1024
    sha1 = hashlib.sha1()
    with open(filename, 'rb') as f:
        while True:
            data = f.read(BUF_SIZE)
            if not data:
                break
            sha1.update(data)
    return sha1.hexdigest()

def download_model():
    """Download pretrained model"""
    # Downloading model from network isn't good for Debian. We need to package
    # the model.
    global pretrained_model
    meta_data = read_model_meta_file(model_meta_file)
    if (meta_data is None):
        logging('Cannot load %s'%(meta_data))
        return None
    # FIXME: using /tmp/ will be in-secure.
    pretrained_model = os.path.join('/','tmp',meta_data['filename'])
    if (os.path.isfile(pretrained_model)):
        sha1 = sha1sum(pretrained_model)
        if (sha1 != meta_data['sha1sum']):
            logging('Model %s SHA1 is not equal to %s'%(pretrained_model, meta_data['sha1sum']))
            pretrained_model = None
            return None
        else:
            logging('Model already exists')
            pass
    else:
        logging('Downloading model file from %s'%(meta_data['url']))
        urllib.request.urlretrieve(meta_data['url'], pretrained_model)
        logging('Checking SHA1...')
        sha1 = sha1sum(pretrained_model)
        if (sha1 != meta_data['sha1sum']):
            logging('Model %s SHA1 is not equal to %s'%(pretrained_model, meta_data['sha1sum']))
            pretrained_model = None
            return None
        else:
            logging('Model downloaded')
            pass
    return None

def download_label():
    """Download label file"""
    # Using the scripts inside caffe Debian package to download label file.
    # This could also be wrong. Why we don't package the label file?
    global label_file
    if (os.path.isfile(label_file)):
        logging("Label file exists");
        pass
    else:
        logging("Label file not exists. Downloading...");
        tmpdir = tempfile.mkdtemp()
        s1 = shutil.copy2(os.path.join('/', 'usr', 'share', 'doc', 'caffe-doc',
                                       'data', 'ilsvrc12',
                                       'get_ilsvrc_aux.sh'),
                          tmpdir)
        os.system('sh \'%s\''%(s1));
        # FIXME: using /tmp/ will be in-secure.
        shutil.copy2(os.path.join(tmpdir, 'synset_words.txt'), '/tmp')

def create_classifier(pretrained_model):
    """Creates a model from saved caffemodel file and returns a classifier."""
    # Creates model from saved .caffemodel.

    # The following file are shipped inside caffe-doc Debian package
    model_def = os.path.join("/", "usr", "share", "doc", "caffe-doc",
                             "models","bvlc_reference_caffenet",
                             "deploy.prototxt")
    image_dims = [ 256, 256 ]
    # The following file are shipped inside python3-caffe-cpu Debian package
    mean = np.load(os.path.join('/', 'usr', 'lib', 'python3',
                                'dist-packages', 'caffe', 'imagenet',
                                'ilsvrc_2012_mean.npy'))
    channel_swap = [2, 1, 0]
    raw_scale = 255.0

    caffe.set_mode_cpu()
    classifier = caffe.Classifier(model_def, pretrained_model,
                                  image_dims=image_dims, mean=mean,
                                  raw_scale=raw_scale,
                                  channel_swap=channel_swap)
    return classifier

def server(labels):
    """Infinite loop serving inference requests"""
    global image_queue, sess

    logging(threading.current_thread().getName(), "is running")

    while True:
        input_name = image_queue.get()
        if (input_name.endswith('npy')):
            inputs = np.load(input_name)
        else:
            inputs = [caffe.io.load_image(input_name)]

        predictions = caffe_classifier.predict(inputs, False)
        # make tuples
        predictions_list = predictions[0].tolist()
        data = zip(predictions_list, caffe_labels)
        output_name = input_name+'.txt'
        output_done_name = output_name+'.done'
        output = open(output_name, 'wt')
        for d in sorted(data, reverse=True):
            human_string = d[1]
            score = d[0]
            print("%s (score = %.5f)" % (human_string, score), file=output)
            if (score < 0.00001):
                break
        output.close()
        touch(output_done_name)
        logging(input_name, " classified!")

class EventHandler(PatternMatchingEventHandler):
    def process(self, event):
        """
        event.event_type
            'modified' | 'created' | 'moved' | 'deleted'
        event.is_directory
            True | False
        event.src_path
            path/to/observed/file
        """
        # the file will be processed there
        global image_queue

        _msg = event.src_path
        image_queue.put(_msg.rstrip('.done'))
        os.remove(_msg)
        logging(_msg, event.event_type)

    # ignore all other types of events except 'modified'
    def on_created(self, event):
        self.process(event)

if __name__ == '__main__':

    pid = str(os.getpid())
    pidfile = "/tmp/classify_server.pid"

    if os.path.isfile(pidfile):
        logging("%s already exists, exiting" % pidfile)
        sys.exit(1)

    with open(pidfile, 'w') as f:
        f.write(pid)

    # Please read /usr/share/doc/caffe-doc/models/bvlc_reference_caffenet/readme.md
    download_model()
    download_label()
    caffe_labels = load_labels(label_file)
    caffe_classifier = create_classifier(pretrained_model)

    # workaround the issue that SIGINT cannot be received (fork a child to
    # avoid blocking the main process in Thread.join()
    child_pid = os.fork()
    if child_pid == 0:
        # child
        # observer handles event in a different thread
        observer = Observer()
        observer.schedule(EventHandler(['*.jpg.done']), path=image_dir)
        observer.start()

        # Create a server thread for each CPU core
        cpu_count = multiprocessing.cpu_count()
        for i in range(1):
            threads.append(threading.Thread(target=server,
                                            name='Server thread %d' % i,
                                            args=({},)))
        for t in threads: t.start()
        for t in threads: t.join()
    else:
        # parent
        try:
            os.wait()
        except KeyboardInterrupt:
            os.kill(child_pid, signal.SIGKILL)
            os.unlink(pidfile)



#!/usr/bin/python3
#
# Copyright 2018 DT42
#
# This file is part of BerryNet.
#
# BerryNet is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BerryNet is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with BerryNet.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import json
import logging
import time

from datetime import datetime

import cv2

from berrynet import logger
from berrynet.comm import Communicator
from berrynet.comm import payload


def parse_args():
    ap = argparse.ArgumentParser()
    ap.add_argument(
        '--mode',
        default='stream',
        help='Camera creates frame(s) from stream or file. (default: stream)'
    )
    ap.add_argument(
        '--stream-src',
        type=str,
        default='0',
        help=('Camera stream source. '
              'It can be device node ID or RTSP URL. '
              '(default: 0)')
    )
    ap.add_argument(
        '--fps',
        type=float,
        default=1,
        help='Frame per second in streaming mode. (default: 1)'
    )
    ap.add_argument(
        '--filepath',
        default='',
        help='Input image path in file mode. (default: empty)'
    )
    ap.add_argument(
        '--broker-ip',
        default='localhost',
        help='MQTT broker IP.'
    )
    ap.add_argument(
        '--broker-port',
        default=1883,
        type=int,
        help='MQTT broker port.'
    )
    ap.add_argument('--topic',
        default='berrynet/data/rgbimage',
        help='The topic to send the captured frames.'
    )
    ap.add_argument('--display',
        action='store_true',
        help=('Open a window and display the sent out frames. '
              'This argument is only effective in stream mode.')
    )
    ap.add_argument('--hash',
        action='store_true',
        help='Add md5sum of a captured frame into the result.'
    )
    ap.add_argument('--meta',
        type=str,
        default='{}',
        help='Metadata field for stringified JSON data.'
    )
    ap.add_argument('--debug',
        action='store_true',
        help='Debug mode toggle'
    )
    return vars(ap.parse_args())


def main():
    args = parse_args()
    if args['debug']:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)

    comm_config = {
        'subscribe': {},
        'broker': {
            'address': args['broker_ip'],
            'port': args['broker_port']
        }
    }
    comm = Communicator(comm_config, debug=True)

    duration = lambda t: (datetime.now() - t).microseconds / 1000

    metadata = json.loads(args.get('meta', '{}'))

    if args['mode'] == 'stream':
        counter = 0
        fail_counter = 0

        # Check input stream source
        if args['stream_src'].isdigit():
            # source is a physically connected camera
            stream_source = int(args['stream_src'])
        else:
            # source is an IP camera
            stream_source = args['stream_src']
        capture = cv2.VideoCapture(stream_source)
        cam_fps = capture.get(cv2.CAP_PROP_FPS)
        if cam_fps > 30 or cam_fps < 1:
            logger.warn('Camera FPS is {} (>30 or <1). Set it to 30.'.format(cam_fps))
            cam_fps = 30
        out_fps = args['fps']
        interval = int(cam_fps / out_fps)

        # warmup
        #t_warmup_start = time.time()
        #t_warmup_now = time.time()
        #warmup_counter = 0
        #while t_warmup_now - t_warmup_start < 1:
        #    capture.read()
        #    warmup_counter += 1
        #    t_warmup_now = time.time()

        logger.debug('===== VideoCapture Information =====')
        if stream_source.isdigit():
            stream_source_uri = '/dev/video{}'.format(stream_source)
        else:
            stream_source_uri = stream_source
        logger.debug('Stream Source: {}'.format(stream_source_uri))
        logger.debug('Camera FPS: {}'.format(cam_fps))
        logger.debug('Output FPS: {}'.format(out_fps))
        logger.debug('Interval: {}'.format(interval))
        logger.debug('Send MQTT Topic: {}'.format(args['topic']))
        #logger.debug('Warmup Counter: {}'.format(warmup_counter))
        logger.debug('====================================')

        while True:
            status, im = capture.read()

            # To verify whether the input source is alive, you should use the
            # return value of capture.read(). It will not work by capturing
            # exception of a capture instance, or by checking the return value
            # of capture.isOpened().
            #
            # Two reasons:
            # 1. If a dead stream is alive again, capture will not notify
            #    that input source is dead.
            #
            # 2. If you check capture.isOpened(), it will keep retruning
            #    True if a stream is dead afterward. So you can not use
            #    the capture return value (capture status) to determine
            #    whether a stream is alive or not.
            if (status is True):
                counter += 1
                if counter == interval:
                    logger.debug('Drop frames: {}'.format(counter-1))
                    counter = 0

                    # Open a window and display the ready-to-send frame.
                    # This is useful for development and debugging.
                    if args['display']:
                        cv2.imshow('Frame', im)
                        cv2.waitKey(1)

                    t = datetime.now()
                    retval, jpg_bytes = cv2.imencode('.jpg', im)
                    mqtt_payload = payload.serialize_jpg(jpg_bytes, args['hash'], metadata)
                    comm.send(args['topic'], mqtt_payload)
                    logger.debug('send: {} ms'.format(duration(t)))
                else:
                    pass
            else:
                fail_counter += 1
                logger.critical('ERROR: Failure #{} happened when reading frame'.format(fail_counter))

                # Re-create capture.
                capture.release()
                logger.critical('Re-create a capture and reconnect to {} after 5s'.format(stream_source))
                time.sleep(5)
                capture = cv2.VideoCapture(stream_source)
    elif args['mode'] == 'file':
        # Prepare MQTT payload
        im = cv2.imread(args['filepath'])
        retval, jpg_bytes = cv2.imencode('.jpg', im)

        t = datetime.now()
        mqtt_payload = payload.serialize_jpg(jpg_bytes, args['hash'], metadata)
        logger.debug('payload: {} ms'.format(duration(t)))
        logger.debug('payload size: {}'.format(len(mqtt_payload)))

        # Client publishes payload
        t = datetime.now()
        comm.send(args['topic'], mqtt_payload)
        logger.debug('mqtt.publish: {} ms'.format(duration(t)))
        logger.debug('publish at {}'.format(datetime.now().isoformat()))
    else:
        logger.error('User assigned unknown mode {}'.format(args['mode']))


if __name__ == '__main__':
    main()



# Copyright 2018 DT42
#
# This file is part of BerryNet.
#
# BerryNet is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BerryNet is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with BerryNet.  If not, see <http://www.gnu.org/licenses/>.

"""Pipeline service with relay engine (default engine).
"""

import cv2
import argparse
import json
import logging
import os
import sys
import time

from datetime import datetime

import numpy as np

from berrynet import logger
from berrynet.bndyda.launcher_berrynet import BerryNetPipelineLauncher
from berrynet.comm import payload
from berrynet.engine import DLEngine
from berrynet.service import EngineService
from dyda_utils import tools


class PipelineEngine(DLEngine):
    def __init__(self, config,
                 dyda_config_path='',
                 warmup_size=(480, 640, 3),
                 disable_warmup=False,
                 benchmark=False,
                 verbosity=0):
        self.launcher = BerryNetPipelineLauncher(
            config,
            dyda_config_path=dyda_config_path,
            verbosity=verbosity, benchmark=benchmark)
        self.pipeline_config = tools.parse_json(config)
        if not disable_warmup:
            self.warmup(shape=warmup_size)

    # def process_input(self, tensor):
    #     return tensor

    def inference(self, tensor, meta={}, base_name=None):
        """
        Args:
            tensor: Image data in BGR format (numpy array)

        Returns:
            Dictionary following generic inference spec and
            pipeline spec (by project).
        """
        return self.launcher.run(tensor, meta=meta, base_name=base_name)
        # return {
        #     'annotations': {
        #         'label': 'dt42',
        #         'confidence': 0.99
        #     }
        # }

    def process_output(self, output):
        return output

    def get_dl_component_config(self, pipeline_config):
        """Get pipeline def list containing only DL components

        Args:
            pipeline_config: pipeline config JSON object

        Returns:
            List of DL components definitions
        """
        dl_comp_config = []
        try:
            pipeline_def = pipeline_config['pipeline_def']
        except KeyError:
            logger.warning('Invalid pipeline config')
            pipeline_def = []
        for comp_config in pipeline_def:
            if ('classifier' in comp_config['name'] or
                    'detector' in comp_config['name']):
                dl_comp_config.append(comp_config)
        return dl_comp_config

    def warmup(self, shape=(480, 640, 3), iteration=5):
        """Warmup pipeline engine

        Use all-zero numpy array as input to warmup pipeline engine.

        Args:
            meta: Metadata of image data
            shape: Warmup image shape in (w, h, c) format
            iteration: How many times to feed in warmup image

        Returns:
            N/A
        """
        logger.debug('Warmup shape: {}'.format(shape))
        input_data = [np.zeros(shape=shape, dtype=np.uint8)] * iteration

        # FIXME: get engines programatically
        dl_comp_config = self.get_dl_component_config(self.pipeline_config)
        for comp_config in dl_comp_config:
            t_start = time.time()

            comp_name = comp_config['name']
            inst = self.launcher.pipeline.pipeline[comp_name]['instance']
            inst.input_data = input_data
            inst.main_process()

            t_duration = time.time() - t_start
            logger.debug('Warmup {0} costs {1} sec'.format(comp_name,
                                                           t_duration))


def duration(t):
    return (datetime.now() - t).microseconds / 1000


class PipelineDummyEngine(DLEngine):
    def inference(self, tensor, meta={}):
        output = None
        return output


class PipelineService(EngineService):
    def __init__(self, service_name, engine, comm_config,
                 pid=None,
                 pipeline_config_path=None,
                 disable_engine=False,
                 disable_warmup=False,
                 warmup_size=(480, 640, 3)):
        super().__init__(service_name,
                         engine,
                         comm_config)

        self.pipeline_config_path = pipeline_config_path
        self.dyda_config_path = ''
        self.warmup_size = warmup_size

        self.disable_engine = disable_engine

        if not os.path.exists('/tmp/dlbox-pipeline'):
            os.mkdir('/tmp/dlbox-pipeline')
        self.counter = 0
        self.pid = pid

        logger.debug('Pipeline result topic: berrynet/engine/pipeline/result')

    def inference(self, pl):
        logger.debug('Disable engine: {}'.format(self.disable_engine))
        if self.disable_engine:
            self.dummy_inference(pl)
        else:
            self.dl_inference(pl)

    def dl_inference(self, pl):
        def empty_inference_result(count):
            return [
                {
                    'channel': i,
                    'annotations': []
                }
                for i in range(count)]

        t = datetime.now()
        base_name = None
        logger.debug('counter #{}'.format(self.counter))
        logger.debug('payload size: {}'.format(len(pl)))
        logger.debug('payload type: {}'.format(type(pl)))
        # Unify the type of input payload to a list, so that
        # bnpipeline can process the input in the same way.
        #
        # If the payload is
        #     - a list of items: keep the list
        #     - a single item: convert to a list with an item
        mqtt_payload = payload.deserialize_payload(pl.decode('utf-8'))
        if isinstance(mqtt_payload, list):
            jpg_json = mqtt_payload
        else:
            jpg_json = [mqtt_payload]
            logger.info('Convert input type from {0} to {1}'.format(
                type(mqtt_payload),
                type(jpg_json)))

        jpg_bytes_list = [
            payload.destringify_jpg(img['bytes']) for img in jpg_json]
        metas = [img.get('meta', {}) for img in jpg_json]
        logger.debug('destringify_jpg: {} ms'.format(duration(t)))

        t = datetime.now()
        bgr_arrays = [
            payload.jpg2bgr(jpg_bytes) for jpg_bytes in jpg_bytes_list]
        logger.debug('jpg2bgr: {} ms'.format(duration(t)))

        t = datetime.now()
        # FIXME: Galaxy pipeline may or may not use a list as input, so we
        # check the length here and then choose whether to send a list or not.
        # We may drop it when Galaxy Pipline unite their input.
        if len(bgr_arrays) > 1:
            image_data = self.engine.process_input(bgr_arrays)
        else:
            image_data = self.engine.process_input(bgr_arrays[0])
        # FIXME: Galaxy pipeline doesn't support multiple metadata for multiple
        # images at the moment (which will be needed), so we provide the first
        # metadata here. This commit should be revert when Galaxy pipeline
        # support it: https://gitlab.com/DT42/galaxy42/dt42-trainer/issues/120
        meta_data = metas[0]

        try:
            logger.debug(meta_data)
            output = self.engine.inference(image_data,
                                           meta=meta_data,
                                           base_name=base_name)
            model_outputs = self.engine.process_output(output)
        except IndexError as e:
            # FIXME: workaround for pipeline
            # Pipeline throw IndexError when there's no results, see:
            # https://gitlab.com/DT42/galaxy42/dt42-trainer/issues/86
            # So we catch the exeception, and produce a dummy result
            # to hook. This workaround should be removed after the issue
            # has been fixed.
            model_outputs = empty_inference_result(len(jpg_json))
            logger.warning(('inference results are empty because '
                            'pipeline raised IndexError'))

        if model_outputs is None:
            model_outputs = empty_inference_result(len(jpg_json))
            logger.warning(('inference results are empty because '
                            'severe error happened in pipeline'))

        logger.debug('Result: {}'.format(model_outputs))
        logger.debug('Classification takes {} ms'.format(duration(t)))

        # self.engine.cache_data('model_output', model_outputs)
        # self.engine.cache_data('model_output_filepath', output_name)
        # self.engine.save_cache()

        self.send_result(self.generalize_result(jpg_json, model_outputs))

        self.counter += 1

    def dummy_inference(self, pl):
        logger.debug('dummy_inference is called')

    def switch_mode(self, pl):
        """Switch pipeline service between inference and non-inference modes

        If Pipeline service receives berrynet/data/mode topic with
        "inference" in payload, service will switch to inference mode;

        If "idle" or "learning" in payload, service will switch to
        non-inference mode.

        Pipeline service will create pipeline engine and
        listen to specified topics only in inference mode.

        Args:
            pl: MQTT message payload
                valid value: {'inference', 'idle', 'learning'}

        Returns:
            N/A
        """
        mode = pl.decode('utf-8')
        if mode == 'inference':
            self.disable_engine = False
            self.engine = PipelineEngine(
                              self.pipeline_config_path,
                              dyda_config_path=self.dyda_config_path,
                              disable_warmup=self.disable_warmup,
                              warmup_size=self.warmup_size)
        else:
            self.disable_engine = True
            self.engine = PipelineDummyEngine()

    def deploy(self, pl):
        """Deploy newly retrained model for pipeline engine

        New dyda config filepath is in the payload.

        Args:
            pl: MQTT message payload w/ new dyda config filepath.

        Returns:
            N/A
        """
        dyda_config_path = pl.decode('utf-8')
        self.dyda_config_path = dyda_config_path
        self.comm.send('berrynet/data/deployed', '')
        logger.info(('New model has been deployed, '
                     'dyda config: {}'.format(self.dyda_config_path)))

    def generalize_result(self, eng_input, eng_output):
        # Pipeline returns None if any error happened
        if eng_output is None:
            eng_output = {}

        # If pipeline generate multiple outputs simultaneously
        #
        # In this case, the format of engine output is
        #
        #     {
        #         'annotations': {...},
        #         'bytes': '...'
        #     }
        if all(key in eng_output.keys() for key in ['annotations', 'bytes']):
            logger.debug('Pipeline output type: multiple')
            logger.debug('eng_input type = {0}, len = {1}'.format(type(eng_input), len(eng_input)))

            # FIXME: Re-cap why eng_input is a list and only contains 1 item.
            eng_input = eng_input[0]

            try:
                eng_input['annotations'] = eng_output['annotations']
                logger.debug('output image type: {0}, len: {1}'.format(type(eng_output['bytes']),
                                                                       len(eng_output['bytes'])))
                pipeline_img = eng_output['bytes'][0]
                retval, jpg_bytes = cv2.imencode('.jpg', pipeline_img)
                eng_input['bytes'] = payload.stringify_jpg(jpg_bytes)
            except Exception as e:
                logger.critical(e)
        else:
            logger.debug('Pipeline output type: simple')

            # FIXME: Workaround for spec incompatibility
            # DLBox spec use 'image_blob', but BerryNet use 'bytes', so we have to
            # do a convert here
            if isinstance(eng_output, list):
                inf_output = eng_output[0]
            else:
                inf_output = eng_output
            if len(eng_input) > 1:
                for i in range(len(eng_input)):
                    try:
                        retval, jpg_bytes = cv2.imencode('.jpg', inf_output)
                        eng_input[i]['bytes'] = payload.stringify_jpg(jpg_bytes)
                        #eng_input[i].pop('bytes')
                    except Exception as e:
                        print(e)

            else:
                try:
                    eng_input, = eng_input
                    retval, jpg_bytes = cv2.imencode('.jpg', inf_output)
                    eng_input['bytes'] = payload.stringify_jpg(jpg_bytes)
                    #eng_input.pop('bytes')
                except Exception as e:
                    print(e)

        return eng_input

    def send_result(self, generalized_result):
        # NOTE: There are numpy float in pipeline output, so we use
        # tools.dump_json instead of payload.serialize_payload
        if self.pid is None:
            self.comm.send(
                'berrynet/engine/pipeline/result',
                tools.dump_json(generalized_result))
        else:
            self.comm.send(
                'berrynet/engine/pipeline/result/{}'.format(self.pid),
                tools.dump_json(generalized_result))


def parse_args():
    ap = argparse.ArgumentParser()
    ap.add_argument('--pipeline-id',
                    help=('Indicate pipeline ID which will be attached '
                          'to result topic (optional)'))
    ap.add_argument('--pipeline-config',
                    help='File contains the definition '
                         'of pipeline application')
    ap.add_argument('--debug',
                    action='store_true',
                    help='Debug mode toggle')
    ap.add_argument('--benchmark',
                    action='store_true',
                    help='Benchmark mode toggle')
    ap.add_argument('--broker-ip',
                    default='localhost',
                    help='MQTT broker IP')
    ap.add_argument('--topic-config',
                    default=None,
                    help='Path of the MQTT topic subscription JSON.')
    ap.add_argument('--topic',
                    nargs=2,
                    action='append',
                    default=None,
                    help=('Two params in "<topic> <handler>" format. '
                          'It can be declared multiple times.'))
    ap.add_argument('--disable-engine',
                    action='store_true',
                    help='Service disable engine initially')
    ap.add_argument('--disable-warmup',
                    action='store_true',
                    help='Skip warming up pipeline by black image')
    ap.add_argument('-v', '--verbosity',
                    action='count', default=0,
                    help='Output verbosity')
    ap.add_argument('-w', '--warmup-size',
                    nargs=2,
                    type=int,
                    default=(640, 480),
                    help='Warmup image\'s size, in format "w h", '
                         'e.g., "640 480"')
    return vars(ap.parse_args())


def main():
    # Process CLI arguments
    args = parse_args()
    if args['debug']:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)

    if args['topic_config']:
        with open(args['topic_config']) as f:
            topic_config = json.load(f)
    else:
        topic_config = {}
    topic_config['berrynet/data/mode'] = 'self.switch_mode'
    topic_config['berrynet/data/deploy'] = 'self.deploy'

    if args['topic'] is not None:
        for t, h in args['topic']:
            topic_config[t] = h

    w, h = args['warmup_size']
    # Setup pipeline service
    if args['disable_engine']:
        eng = PipelineDummyEngine()
    else:
        eng = PipelineEngine(args['pipeline_config'],
                             disable_warmup=args['disable_warmup'],
                             verbosity=args['verbosity'],
                             benchmark=args['benchmark'],
                             warmup_size=(h, w, 3))
    comm_config = {
        'subscribe': topic_config,
        'broker': {
            'address': args['broker_ip'],
            'port': 1883
        }
    }
    engine_service = PipelineService(
        'pipeline service',
        eng,
        comm_config,
        pid=args['pipeline_id'],
        pipeline_config_path=args['pipeline_config'],
        disable_engine=args['disable_engine'],
        disable_warmup=args['disable_warmup'],
        warmup_size=(h, w, 3))

    engine_service.run(args)


if __name__ == '__main__':
    main()



# Copyright 2019 DT42
#
# This file is part of BerryNet.
#
# BerryNet is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BerryNet is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with BerryNet.  If not, see <http://www.gnu.org/licenses/>.

"""Engine service is a bridge between incoming data and inference engine.
"""

import argparse
import logging
import time

from berrynet import logger
from berrynet.comm import payload
from berrynet.dlmodelmgr import DLModelManager
from berrynet.engine.tflite_engine import TFLiteClassifierEngine
from berrynet.engine.tflite_engine import TFLiteDetectorEngine
from berrynet.service import EngineService
from berrynet.utils import draw_bb
from berrynet.utils import generate_class_color


class TFLiteClassifierService(EngineService):
    def __init__(self, service_name, engine, comm_config, draw=False):
        super(TFLiteClassifierService, self).__init__(service_name,
                                                      engine,
                                                      comm_config)
        self.draw = draw

    def inference(self, pl):
        t0 = time.time()
        logger.debug('payload size: {}'.format(len(pl)))
        logger.debug('payload type: {}'.format(type(pl)))
        jpg_json = payload.deserialize_payload(pl.decode('utf-8'))
        jpg_bytes = payload.destringify_jpg(jpg_json['bytes'])
        logger.debug('destringify_jpg: {} ms'.format(time.time() - t0))

        t1 = time.time()
        bgr_array = payload.jpg2bgr(jpg_bytes)
        logger.debug('jpg2bgr: {} ms'.format(time.time() - t1))

        t2 = time.time()
        image_data = self.engine.process_input(bgr_array)
        logger.debug('Input processing takes {} ms'.format(time.time() - t2))

        t3 = time.time()
        output = self.engine.inference(image_data)
        model_outputs = self.engine.process_output(output)
        logger.debug('Result: {}'.format(model_outputs))
        logger.debug('Classification takes {} ms'.format(time.time() - t3))

        classes = self.engine.classes
        labels = self.engine.labels

        logger.debug('draw = {}'.format(self.draw))
        if self.draw is False:
            self.result_hook(self.generalize_result(jpg_json, model_outputs))
        else:
            self.result_hook(
                draw_label(bgr_array,
                           self.generalize_result(jpg_json, model_outputs),
                           color,
                           labels))

    def result_hook(self, generalized_result):
        logger.debug('result_hook, annotations: {}'.format(generalized_result['annotations']))
        self.comm.send('berrynet/engine/tfliteclassifier/result',
                       payload.serialize_payload(generalized_result))


class TFLiteDetectorService(EngineService):
    def __init__(self, service_name, engine, comm_config, draw=False):
        super(TFLiteDetectorService, self).__init__(service_name,
                                                   engine,
                                                   comm_config)
        self.draw = draw

    def inference(self, pl):
        t0 = time.time()
        logger.debug('payload size: {}'.format(len(pl)))
        logger.debug('payload type: {}'.format(type(pl)))
        jpg_json = payload.deserialize_payload(pl.decode('utf-8'))
        jpg_bytes = payload.destringify_jpg(jpg_json['bytes'])
        logger.debug('destringify_jpg: {} ms'.format(time.time() - t0))

        t1 = time.time()
        bgr_array = payload.jpg2bgr(jpg_bytes)
        logger.debug('jpg2bgr: {} ms'.format(time.time() - t1))

        t2 = time.time()
        image_data = self.engine.process_input(bgr_array)
        output = self.engine.inference(image_data)
        model_outputs = self.engine.process_output(output)
        logger.debug('Result: {}'.format(model_outputs))
        logger.debug('Detection takes {} ms'.format(time.time() - t2))

        classes = self.engine.classes
        labels = self.engine.labels

        logger.debug('draw = {}'.format(self.draw))
        if self.draw is False:
            self.result_hook(self.generalize_result(jpg_json, model_outputs))
        else:
            self.result_hook(
                draw_bb(bgr_array,
                        self.generalize_result(jpg_json, model_outputs),
                        generate_class_color(class_num=classes),
                        labels))

    def result_hook(self, generalized_result):
        logger.debug('result_hook, annotations: {}'.format(generalized_result['annotations']))
        self.comm.send('berrynet/engine/tflitedetector/result',
                       payload.serialize_payload(generalized_result))


def parse_args():
    ap = argparse.ArgumentParser()
    ap.add_argument(
        '-s', '--service',
        help=('Classifier or Detector service. '
              'classifier, or detector is acceptable. '
              '(classifier by default)'),
        default='classifier',
        type=str)
    ap.add_argument(
        '--service_name',
        default='tflite_classifier',
        help='Human-readable service name for service management.')
    ap.add_argument(
        '-m', '--model',
        help='Model file path')
    ap.add_argument(
        '-l', '--label',
        help='Label file path')
    ap.add_argument(
        '-p', '--model_package',
        default='',
        help='Model package name. Find model and label file paths automatically.')
    ap.add_argument(
        '--top_k',
        help='Display top K classification results.',
        default=3,
        type=int)
    ap.add_argument(
        '--num_threads',
        default=1,
        help="Number of threads for running inference.",
        type=int)
    ap.add_argument(
        '--draw',
        action='store_true',
        help='Draw bounding boxes on image in result')
    ap.add_argument(
        '--debug',
        action='store_true',
        help='Debug mode toggle')
    return vars(ap.parse_args())


def main():
    # Test TFLite engines
    args = parse_args()
    if args['debug']:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)
    if args['model_package'] != '':
        dlmm = DLModelManager()
        meta = dlmm.get_model_meta(args['model_package'])
        args['model'] = meta['model']
        args['label'] = meta['label']
    logger.debug('model filepath: ' + args['model'])
    logger.debug('label filepath: ' + args['label'])

    comm_config = {
        'subscribe': {},
        'broker': {
            'address': 'localhost',
            'port': 1883
        }
    }

    if args['service'] == 'classifier':
        engine = TFLiteClassifierEngine(
                     model = args['model'],
                     labels = args['label'],
                     top_k = args['top_k'],
                     num_threads = args['num_threads'])
        service_functor = TFLiteClassifierService
    elif args['service'] == 'detector':
        engine = TFLiteDetectorEngine(
                     model = args['model'],
                     labels = args['label'],
                     num_threads = args['num_threads'])
        service_functor = TFLiteDetectorService
    else:
        raise Exception('Illegal service {}, it should be '
                        'classifier or detector'.format(args['service']))

    engine_service = service_functor(args['service_name'],
                                     engine,
                                     comm_config,
                                     draw=args['draw'])
    engine_service.run(args)


if __name__ == '__main__':
    main()



import subprocess

from setuptools import setup
from setuptools import find_packages


with open("README.md", "r") as f:
    long_description = f.read()


setup(
    name='berrynet',
    version='v3.10.2',
    description='Deep learning gateway on Raspberry Pi and other edge devices.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/DT42/BerryNet',
    author='DT42 Inc.',
    author_email='berrynet@dt42.io',
    license='GPLv3',
    # https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    keywords=['wheels'],
    packages=find_packages(exclude=['tests']),
    install_requires=[
        'logzero',
        'paho-mqtt',
        'PyOpenGL',
        'PyOpenGL-accelerate',
    ],
    # Without TF: pip3 install berrynet
    # TF CPU mode: pip3 install berrynet[tf]
    # TF GPU mode: pip3 install berrynet[tf_gpu]
    # TF GPU mode and Wheel-version OpenCV: pip3 install berrynet[tf_gpu, opencv]
    extra_require={
        'tf': ['tensorflow>=1.2.1'],
        'tf_gpu': ['tensorflow-gpu>=1.2.1'],
        'opencv': ['opencv-python']
    },
    python_requires='>=3',  # recognized by pip >= 9.0.0
    # The "Including Data Files" session
    # https://setuptools.readthedocs.io/en/latest/setuptools.html
    #
    # Warning: package_data only looks for module's top level
    #package_data={
    #},
    #data_files=[
    #    ('docs', ['docs/cheatsheet.txt', 'docs/references.txt']),
    #],
    #scripts=[
    #    'bin/purescript.sh'
    #],
    # The "Automatic Script Creation" session
    # https://setuptools.readthedocs.io/en/latest/setuptools.html
    #
    # http://python-packaging.readthedocs.io/en/latest/command-line-scripts.html
    entry_points={
        'console_scripts': [
            'bn_camera=berrynet.client.camera:main',
            'bn_dashboard=berrynet.client.fbdashboard:main',
            'bn_data_collector=berrynet.client.data_collector:main',
            'bn_gmail=berrynet.client.gmail:main',
            'bn_telegram=berrynet.client.telegram_bot:main',
            'bn_tflite=berrynet.service.tflite_service:main',
            'bn_openvino=berrynet.service.openvino_service:main',
            'bn_darknet=berrynet.service.darknet_service:main',
            'bn_pipeline=berrynet.bndyda.bnpipeline:main',
        ]
    },
    #ext_modules=[
    #    Extension(
    #        'hellowheels.lib.clib',
    #        [
    #            'hellowheels/lib/clib.c',
    #            'hellowheels/lib/ext.c'
    #        ]
    #    )
    #],
    #cmdclass={
    #    'install': CustomInstallCommand
    #},
    test_suite='tests'
)



# Copyright 2017 DT42
# 
# This file is part of BerryNet.
# 
# BerryNet is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# BerryNet is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with BerryNet.  If not, see <http://www.gnu.org/licenses/>.

"""Simple image classification server with Inception.

The server monitors image_dir and run inferences on new images added to the
directory. Every image file should come with another empty file with '.done'
suffix to signal readiness. Inference result of a image can be read from the
'.txt' file of that image after '.txt.done' is spotted.

This is an example the server expects clients to do. Note the order.

# cp cat.jpg /run/image_dir
# touch /run/image_dir/cat.jpg.done

Clients should wait for appearance of 'cat.jpg.txt.done' before getting
result from 'cat.jpg.txt'.
"""


from __future__ import print_function
import os
import sys
import time
from caffe2.proto import caffe2_pb2
import numpy as np
import skimage.io
import skimage.transform
import threading
import multiprocessing
import Queue
import signal
from watchdog.observers import Observer
from watchdog.events import PatternMatchingEventHandler
from caffe2.python import core, workspace
import urllib2

image_dir = '/run/image_dir'
image_queue = Queue.Queue()
sess = None
threads = []

def logging(*args):
    print("[%08.3f]" % time.time(), ' '.join(args))

def touch(fname, times=None):
    with open(fname, 'a'):
        os.utime(fname, times)

def crop_center(img,cropx,cropy):
    y,x,c = img.shape
    startx = x//2-(cropx//2)
    starty = y//2-(cropy//2)
    return img[starty:starty+cropy,startx:startx+cropx]

def rescale(img, input_height, input_width):
    aspect = img.shape[1]/float(img.shape[0])
    if(aspect>1):
        # landscape orientation - wide image
        res = int(aspect * input_height)
        imgScaled = skimage.transform.resize(img, (input_width, res))
    if(aspect<1):
        # portrait orientation - tall image
        res = int(input_width/aspect)
        imgScaled = skimage.transform.resize(img, (res, input_height))
    if(aspect == 1):
        imgScaled = skimage.transform.resize(img, (input_width, input_height))
    return imgScaled

def server(labels):
    """Infinite loop serving inference requests"""

    global image_queue, sess
    CAFFE2_ROOT = "/caffe2"
    CAFFE_MODELS = "/caffe2/caffe2/python/models"
    MODEL = 'squeezenet', 'exec_net.pb', 'predict_net.pb', 'ilsvrc_2012_mean.npy', 227
    codes =  "https://gist.githubusercontent.com/aaronmarkham/cd3a6b6ac071eca6f7b4a6e40e6038aa/raw/9edb4038a37da6b5a44c3b5bc52e448ff09bfe5b/alexnet_codes"

    logging(threading.current_thread().getName(), "is running")
    CAFFE2_ROOT = os.path.expanduser(CAFFE2_ROOT)
    CAFFE_MODELS = os.path.expanduser(CAFFE_MODELS)
    MEAN_FILE = os.path.join(CAFFE_MODELS, MODEL[0], MODEL[3])
    if not os.path.exists(MEAN_FILE):
        mean = 128
    else:
        mean = np.load(MEAN_FILE).mean(1).mean(1)
        mean = mean[:, np.newaxis, np.newaxis]
    INPUT_IMAGE_SIZE = MODEL[4]
    INIT_NET = os.path.join(CAFFE_MODELS, MODEL[0], MODEL[1])
    PREDICT_NET = os.path.join(CAFFE_MODELS, MODEL[0], MODEL[2])

    while True:
        input_name = image_queue.get()
        img = skimage.img_as_float(skimage.io.imread(input_name)).astype(np.float32)
        img = rescale(img, INPUT_IMAGE_SIZE, INPUT_IMAGE_SIZE)
        img = crop_center(img, INPUT_IMAGE_SIZE, INPUT_IMAGE_SIZE)
            
        img = img.swapaxes(1, 2).swapaxes(0, 1)
        img = img[(2, 1, 0), :, :]
        img = img * 255 - mean
        img = img[np.newaxis, :, :, :].astype(np.float32)

        with open(INIT_NET, 'rb') as f:
            init_net = f.read()
        with open(PREDICT_NET, 'rb') as f:
            predict_net = f.read()
            
        p = workspace.Predictor(init_net, predict_net)
            
        # run the net and return prediction
        results = p.run([img])
        results = np.asarray(results)
        results = np.delete(results, 1)
        index = 0
        highest = 0
        arr = np.empty((0,2), dtype=object)
        arr[:,0] = int(10)
        arr[:,1:] = float(10)
        for i, r in enumerate(results):
            # imagenet index begins with 1!
            i=i+1
            arr = np.append(arr, np.array([[i,r]]), axis=0)
            if (r > highest):
                highest = r
                index = i
        response = urllib2.urlopen(codes)
        output_name = input_name+'.txt'
        output_done_name = output_name+'.done'
        output = open(output_name, 'w')
        for line in response:
            code, result = line.partition(":")[::2]
            if (code.strip() == str(index)):
                human_string = result.strip()[1:-2]
                score = highest
                print("%s (score = %.5f)" % (human_string, score), file=output)
        output.close()
        touch(output_done_name)
        logging(input_name, " classified!")


class EventHandler(PatternMatchingEventHandler):
    def process(self, event):
        """
        event.event_type
            'modified' | 'created' | 'moved' | 'deleted'
        event.is_directory
            True | False
        event.src_path
            path/to/observed/file
        """
        # the file will be processed there
        global image_queue

        _msg = event.src_path
        image_queue.put(_msg.rstrip('.done'))
        os.remove(_msg)
        logging(_msg, event.event_type)

    # ignore all other types of events except 'modified'
    def on_created(self, event):
        self.process(event)


def main(_):
    """Called by Tensorflow"""
 

    # Create a server thread for each CPU core
    cpu_count = multiprocessing.cpu_count()
    for i in xrange(cpu_count/4):
        threads.append(threading.Thread(target=server,
                                        name='Server thread %d' % i,
                                        args=({},)))
    for t in threads: t.start()
    for t in threads: t.join()
 

if __name__ == '__main__':
    global sess, threads

    pid = str(os.getpid())
    pidfile = "/tmp/classify_server.pid"

    if os.path.isfile(pidfile):
        logging("%s already exists, exiting" % pidfile)
        sys.exit(1)

    with open(pidfile, 'w') as f:
        f.write(pid)

    # workaround the issue that SIGINT cannot be received (fork a child to 
    # avoid blocking the main process in Thread.join()
    child_pid = os.fork()
    if child_pid == 0:
        # child
        # observer handles event in a different thread
        observer = Observer()
        observer.schedule(EventHandler(['*.jpg.done']), path=image_dir)
        observer.start()
        # Create a server thread for each CPU core
        cpu_count = multiprocessing.cpu_count()
        for i in xrange(cpu_count/4):
            threads.append(threading.Thread(target=server,
                                            name='Server thread %d' % i,
                                            args=({},)))
        for t in threads: t.start()
        for t in threads: t.join()
    else:
        # parent
        try:
            os.wait()
        except KeyboardInterrupt:
            os.kill(child_pid, signal.SIGKILL)
            os.unlink(pidfile)



import unittest

import cv2

from berrynet.engine.tensorflow_engine import TensorFlowEngine


class TestTensorFlowEngine(unittest.TestCase):
    def test_engine(self):
        model = 'berrynet/engine/inception_v3_2016_08_28_frozen.pb'
        label = 'berrynet/engine/imagenet_slim_labels.txt'
        jpg_filepath = 'berrynet/engine/grace_hopper.jpg'
        input_layer = 'input:0'
        output_layer = 'InceptionV3/Predictions/Reshape_1:0'

        tfe = TensorFlowEngine(model, label, input_layer, output_layer)
        tfe.create()
        rgb_array = cv2.cvtColor(
            cv2.imread(jpg_filepath),
            cv2.COLOR_BGR2RGB)
        tfe.process_output(tfe.inference(tfe.process_input(rgb_array)))
        tfe.process_output(tfe.inference(tfe.process_input(rgb_array)))
        #self.assertEqual('foo'.upper(), 'FOO')

    #def test_isupper(self):
    #    self.assertTrue('FOO'.isupper())
    #    self.assertFalse('Foo'.isupper())

    #def test_split(self):
    #    s = 'hello world'
    #    self.assertEqual(s.split(), ['hello', 'world'])
    #    # check that s.split fails when the separator is not a string
    #    with self.assertRaises(TypeError):
    #        s.split(2)


if __name__ == '__main__':
    unittest.main()



# Copyright 2018 DT42
#
# This file is part of BerryNet.
#
# BerryNet is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BerryNet is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with BerryNet.  If not, see <http://www.gnu.org/licenses/>.

"""Darknet inference engine.
"""

from __future__ import print_function

import argparse
import json
import math
import time

import cv2
import numpy as np

from berrynet import logger
#from berrynet.dlmodelmgr import DLModelManager
from berrynet.engine import DLEngine
from ctypes import *


class BOX(Structure):
    _fields_ = [("x", c_float),
                ("y", c_float),
                ("w", c_float),
                ("h", c_float)]

class IMAGE(Structure):
    _fields_ = [("w", c_int),
                ("h", c_int),
                ("c", c_int),
                ("data", POINTER(c_float))]

class METADATA(Structure):
    _fields_ = [("classes", c_int),
                ("names", POINTER(c_char_p))]


lib = CDLL("/usr/lib/libdarknet.so", RTLD_GLOBAL)
lib.network_width.argtypes = [c_void_p]
lib.network_width.restype = c_int
lib.network_height.argtypes = [c_void_p]
lib.network_height.restype = c_int

predict = lib.network_predict
predict.argtypes = [c_void_p, POINTER(c_float)]
predict.restype = POINTER(c_float)

make_image = lib.make_image
make_image.argtypes = [c_int, c_int, c_int]
make_image.restype = IMAGE

make_boxes = lib.make_boxes
make_boxes.argtypes = [c_void_p]
make_boxes.restype = POINTER(BOX)

free_ptrs = lib.free_ptrs
free_ptrs.argtypes = [POINTER(c_void_p), c_int]

num_boxes = lib.num_boxes
num_boxes.argtypes = [c_void_p]
num_boxes.restype = c_int

make_probs = lib.make_probs
make_probs.argtypes = [c_void_p]
make_probs.restype = POINTER(POINTER(c_float))

reset_rnn = lib.reset_rnn
reset_rnn.argtypes = [c_void_p]

load_net = lib.load_network
load_net.argtypes = [c_char_p, c_char_p, c_int]
load_net.restype = c_void_p

free_image = lib.free_image
free_image.argtypes = [IMAGE]

letterbox_image = lib.letterbox_image
letterbox_image.argtypes = [IMAGE, c_int, c_int]
letterbox_image.restype = IMAGE

load_meta = lib.get_metadata
lib.get_metadata.argtypes = [c_char_p]
lib.get_metadata.restype = METADATA

load_image = lib.load_image_color
load_image.argtypes = [c_char_p, c_int, c_int]
load_image.restype = IMAGE

rgbgr_image = lib.rgbgr_image
rgbgr_image.argtypes = [IMAGE]

predict_image = lib.network_predict_image
predict_image.argtypes = [c_void_p, IMAGE]
predict_image.restype = POINTER(c_float)

network_detect = lib.network_detect
network_detect.argtypes = [c_void_p, IMAGE, c_float, c_float, c_float, POINTER(BOX), POINTER(POINTER(c_float))]


def c_array(ctype, values):
    arr = (ctype*len(values))()
    arr[:] = values
    return arr


def nparray_to_image(arr):
    """Convert nparray to Darknet image struct.
    Args:
        arr: nparray containing source image in BGR color model.

    Returns:
        Darknet image struct, whose data is a C array
        containing flatten image in BGR color model.
    """
    arr = arr.transpose(2,0,1)
    c = arr.shape[0]
    h = arr.shape[1]
    w = arr.shape[2]
    arr = (arr/255.0).flatten()
    data = c_array(c_float, arr)
    im = IMAGE(w, h, c, data)
    rgbgr_image(im)
    return im


def detect_np(net, meta, np_img, thresh=.3, hier_thresh=.5, nms=.45):
    im = nparray_to_image(np_img)
    boxes = make_boxes(net)
    probs = make_probs(net)
    num = num_boxes(net)
    t_start = time.time()
    network_detect(net, im, thresh, hier_thresh, nms, boxes, probs)
    t_end = time.time()
    logger.debug('inference time: {} s'.format(t_end - t_start))
    res = []
    for j in range(num):
        for i in range(meta.classes):
            if probs[j][i] > 0:
                res.append(
                    {
                        'type': 'detection',
                        'label': meta.names[i].decode('utf-8'),
                        'confidence': probs[j][i],
                        'left': boxes[j].x - (boxes[j].w / 2),
                        'top': boxes[j].y - (boxes[j].h / 2),
                        'right': boxes[j].x + (boxes[j].w / 2),
                        'bottom': boxes[j].y + (boxes[j].h / 2),
                        'id': -1
                    }
                )
    free_ptrs(cast(probs, POINTER(c_void_p)), num)
    return res


class DarknetEngine(DLEngine):
    # FIXME: Get model information by model manager
    def __init__(self, config, model, meta=''):
        super(DarknetEngine, self).__init__()

        self.net = load_net(config, model, 0)
        self.meta = load_meta(meta)
        self.classes = self.meta.classes
        self.labels = [self.meta.names[i].decode('utf-8')
                       for i in range(self.classes)]

        # Warmup
        zero_image = np.zeros(shape=(416, 416, 3), dtype=np.uint8)
        detect_np(self.net, self.meta, zero_image)

    def process_input(self, rgb_array):
        return rgb_array

    def inference(self, tensor):
        return detect_np(self.net, self.meta, tensor)

    def process_output(self, output):
        return {'annotations': output}


if __name__ == '__main__':
    engine = DarknetEngine(
        config=b'/usr/share/dlmodels/tinyyolovoc-20170816/tiny-yolo-voc.cfg',
        model=b'/usr/share/dlmodels/tinyyolovoc-20170816/tiny-yolo-voc.weights',
        meta=b'/usr/share/dlmodels/tinyyolovoc-20170816/voc.data'
    )
    im = cv2.imread('data/dog.jpg')
    for i in range(3):
        r = engine.inference(im)
        print(r)



# Copyright 2017 DT42
#
# This file is part of BerryNet.
#
# BerryNet is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BerryNet is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with BerryNet.  If not, see <http://www.gnu.org/licenses/>.

"""TensorFlow inference engine.
"""

from __future__ import print_function

import argparse
import json

import numpy as np
import caffe

from berrynet import logger
#from berrynet.dlmodelmgr import DLModelManager
from berrynet.engine import DLEngine


class CaffeEngine(DLEngine):
    # FIXME: Get model information by model manager
    def __init__(self, model_def, pretrained_model, mean_file, label, image_dims = [256,256], channel_swap=[2,1,0], raw_scale=255.0, top_k=5):
        super(CaffeEngine, self).__init__()

        # Load model
        caffe.set_mode_cpu()
        self.classifier = caffe.Classifier(model_def, pretrained_model, image_dims=image_dims, mean=mean_file, raw_scale=raw_scale, channel_swap=channel_swap)
        
        # Load labels
        self.labels = [line.rstrip() for line in open(label)]

        self.top_k = top_k

    def create(self):
        pass

    def process_input(self, rgb_array):
        self.inputs = rgb_array
        return self.inputs
    
    def inference(self, tensor):
        self.predictions = self.classifier.predict(self.inputs, False)
        return self.predictions

    def process_output(self, output):
        predictions_list = self.predictions[0].tolist()
        data = zip(predictions_list, caffe_labels)
        processed_output = {'annotations': []}
        i=0
        for d in sorted(data, reverse=True):
            human_string = d[1]
            score = d[0]
            anno = {
                'type': 'classification',
                'label': human_string,
                'confidence': score
            }
            processed_output['annotations'].append(anno)
            i = i + 1
            if (i >= self.top_k):
                break
        return processed_output

    def save_cache(self):
        pass



# Copyright 2018 DT42
#
# This file is part of BerryNet.
#
# BerryNet is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BerryNet is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with BerryNet.  If not, see <http://www.gnu.org/licenses/>.

import base64
import hashlib
import json

from datetime import datetime

import cv2
import numpy as np


def stringify_jpg(jpg_bytes):
    return base64.b64encode(jpg_bytes).decode('utf-8')


def destringify_jpg(stringified_jpg):
    """
    :return: JPEG bytes
    :rtype: bytes
    """
    return base64.b64decode(stringified_jpg.encode('utf-8'))


def jpg2bgr(jpg_bytes):
    """
    :return: BGR bytes
    :rtype: numpy array
    """
    array = np.frombuffer(jpg_bytes, dtype=np.uint8)
    return cv2.imdecode(array, flags=1)


def jpg2rgb(jpg_bytes):
    """
    :return: RGB bytes
    :rtype: numpy array
    """
    return cv2.cvtColor(jpg2bgr(jpg_bytes), cv2.COLOR_BGR2RGB)


def bgr2rgb(bgr_nparray):
    """Convert image nparray from BGR to RGB.

    Args:
        bgr_nparray: Image nparray in BGR color model.

    Returns:
        Image nparray in RGB color model.
    """
    return cv2.cvtColor(bgr_nparray, cv2.COLOR_BGR2RGB)


def rgb2bgr(rgb_nparray):
    """Convert image nparray from RGB to BGR.

    Args:
        rgb_nparray: Image nparray in RGB color model.

    Returns:
        Image nparray in BGR color model.
    """
    return cv2.cvtColor(rgb_nparray, cv2.COLOR_RGB2BGR)


def generate_bytes_md5sum(content_bytes):
    content_b64 = base64.b64encode(content_bytes)
    return hashlib.md5(content_b64).hexdigest()


def serialize_payload(json_object):
    return json.dumps(json_object)


def serialize_jpg(jpg_bytes, md5sum=False, meta={}):
    """Create Serialized JSON object consisting of image bytes and meta

    :param imarray: JPEG bytes
    :type imarray: bytes
    :return: serialized image JSON
    :rtype: string
    """
    obj = {}
    obj['timestamp'] = datetime.now().isoformat()
    obj['bytes'] = stringify_jpg(jpg_bytes)
    obj['meta'] = meta
    if md5sum:
        obj['md5sum'] = generate_bytes_md5sum(jpg_bytes)
    return json.dumps(obj)


def deserialize_payload(payload):
    return json.loads(payload)


#def deserialize_jpg(jpg_json):
#    """Deserialized JSON object created by josnify_image.
#
#    :param string :
#    :return:
#    :rtype:
#    """
#    return json.loads(jpg_json)


if __name__ == '__main__':
    im = cv2.imread('/home/debug/codes/darknet/data/dog.jpg')
    retval, jpg_bytes = cv2.imencode('.jpg', im)

    # size of stringified dog.jpg is 1.33x larger than original
    s_jpg = serialize_jpg(jpg_bytes)
    d_jpg = deserialize_payload(s_jpg)
    # TODO: Can we write JPEG bytes into file directly to prevent
    #       bytes -> numpy array -> decode RGB -> write encoded JPEG
    cv2.imwrite('/tmp/dog.jpg', jpg2bgr(destringify_jpg(d_jpg['bytes'])))



# Copyright 2018 DT42
#
# This file is part of BerryNet.
#
# BerryNet is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BerryNet is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with BerryNet.  If not, see <http://www.gnu.org/licenses/>.

"""Dashboard agent service.
"""

import argparse
import json

from os.path import join as pjoin

from berrynet import logger
from berrynet.comm import Communicator
from berrynet.comm import payload


class DashboardService(object):
    def __init__(self, service_name, comm_config):
        self.service_name = service_name
        self.comm_config = comm_config
        self.comm_config['subscribe']['berrynet/engine/tensorflow/result'] = self.update
        self.comm_config['subscribe']['berrynet/engine/mvclassification/result'] = self.update
        self.comm = Communicator(self.comm_config, debug=True)
        self.basedir = '/usr/local/berrynet/dashboard/www/freeboard'

    def update(self, pl):
        payload_json = payload.deserialize_payload(pl.decode('utf-8'))
        jpg_bytes = payload.destringify_jpg(payload_json['bytes'])
        inference_result = [
            '{0}: {1}<br>'.format(anno['label'], anno['confidence'])
            for anno in payload_json['annotations']
        ]
        logger.debug('inference results: {}'.format(inference_result))

        with open(pjoin(self.basedir, 'snapshot.jpg'), 'wb') as f:
            f.write(jpg_bytes)
        self.comm.send('berrynet/dashboard/snapshot', 'snapshot.jpg')
        self.comm.send('berrynet/dashboard/inferenceResult',
                       json.dumps(inference_result))

    def run(self, args):
        """Infinite loop serving inference requests"""
        self.comm.run()


def parse_args():
    ap = argparse.ArgumentParser()
    ap.add_argument('--service_name', required=True,
                    help='Engine service name used as PID filename')
    return vars(ap.parse_args())


def main():
    args = parse_args()

    comm_config = {
        'subscribe': {},
        'broker': {
            'address': 'localhost',
            'port': 1883
        }
    }
    dashboard_service = DashboardService(args['service_name'],
                                         comm_config)
    dashboard_service.run(args)


if __name__ == '__main__':
    main()



# Copyright 2018 DT42
#
# This file is part of BerryNet.
#
# BerryNet is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BerryNet is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with BerryNet.  If not, see <http://www.gnu.org/licenses/>.

"""Movidius classification inference engine.
"""

from __future__ import print_function

from berrynet.engine import DLEngine
from berrynet.engine import movidius as mv


class MovidiusEngine(DLEngine):
    def __init__(self, model, label):
        super(MovidiusEngine, self).__init__()
        self.mvng = mv.MovidiusNeuralGraph(model, label)

    def process_input(self, tensor):
        return mv.process_inceptionv3_input(tensor)

    def inference(self, tensor):
        return self.mvng.inference(tensor)

    def process_output(self, output):
        return mv.process_inceptionv3_output(
                   output,
                   self.mvng.get_labels())

    def save_cache(self):
        with open(self.cache['model_output_filepath'], 'w') as f:
            for i in self.cache['model_output']:
                print("%s (score = %.5f)" % (i[0], i[1]), file=f)


class MovidiusMobileNetSSDEngine(DLEngine):
    def __init__(self, model, label):
        super(MovidiusMobileNetSSDEngine, self).__init__()
        self.mvng = mv.MovidiusNeuralGraph(model, label)
        self.labels = self.mvng.get_labels()
        self.classes = len(self.labels)

    def process_input(self, tensor):
        self.img_w = tensor.shape[1]
        self.img_h = tensor.shape[0]
        return mv.process_mobilenetssd_input(tensor)

    def inference(self, tensor):
        return self.mvng.inference(tensor)

    def process_output(self, output):
        return mv.process_mobilenetssd_output(
                   output,
                   self.img_w,
                   self.img_h,
                   self.mvng.get_labels())
